/*
   olden.c - convert modern letters to archaic English letters.

   Copyright (C) 2019 Paul Hardy <unifoundry@unifoundry.com>


   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>
#include <stdlib.h>

#include "config.h"
#include "olden.h"


#define MAXSTRING 256

/*
   Substitutions to perform.
*/
#define USE_ALL		0xFFFF
#define USE_ASH		0x01
#define USE_ETHEL	0x02
#define USE_ETH		0x04
#define USE_THORN	0x08
#define USE_TH		0x0C
#define USE_LONG_S	0x10
#define USE_YOGH	0x20
#define USE_WYNN	0x40
#define USE_INSULAR	0xF00

int main (int argc, char *argv[]) {
   int i;  /* loop variable */
   unsigned use_mask = USE_ALL;  /* default = apply all transformations */
   char instring[MAXSTRING];     /* input string to search              */
   int  printeth_help = 0;       /* flag to print help message & exit   */
   char infile[MAXSTRING];       /* name of input file                  */
   char outfile[MAXSTRING];      /* name of output file                 */
   FILE *infp, *outfp;           /* input & output file pointers        */
   int  bom_out=0;               /* =1 to begin with Byte Order Mark    */

   /*
      Definitions for working with a map file of (orig, new) pairs.
   */
   char  map_file[MAXSTRING] = {'\0'};  /* name of map file, if specified */
   int   n_entries;      /* number of entries found in map file */
   int   n_mapped=0;     /* number of entries that were mapped */
   char *back_pat;       /* pointer in string of a "\\" backslash pattern */
   char *map;            /* array of entire map file */
   char **orig_map;      /* index into array of original words*/
   char **new_map;       /* index into array of new words */

   char *mapfile;  /* name of map file; an argument */
   long  nbytes;   /* number of bytes in map file */
   char  linein [MAXSTRING];  /* input line from map file */

   FILE *mapfp;    /* map file pointer */

   int parseth_map (char *map, long nbytes, char **orig_map, char **new_map);
   void replaceth_map (char *sstring, char **orig_map, char **new_map,
                       int n_mapped);

   /*
      Non-map file related function prototypes.
   */
   void replaceth_ae (char *);
   void replaceth_oe (char *);
   void replaceth_gh (char *);
   void replaceth_s  (char *);
   void replaceth_th (char *, unsigned);
   void replaceth_w  (char *);
   void replaceth_ye_old (char *, unsigned);
   void replaceth_insular (char *);
   void replaceth_back (char *);

   void parseth_args (int argc, char *argv[], int *printeth_help,
                      unsigned *use_mask, char *infile, char *outfile,
                      int *bom_out, char *map_file);
   void helpeth_out ();

   infp  = stdin;   /* default value */
   outfp = stdout;  /* default value */
   /*
      Parse the command line arguments.  If none, assign default value of
      USE_ALL to use_mask, and set infile and outfile to stdin and stdout,
      respectively.
   */
   parseth_args (argc, argv,
                 &printeth_help, &use_mask, infile, outfile,
                 &bom_out, map_file);

   /* Unknown option entered; output a help message to stderr and exit. */
   if (printeth_help) {
      helpeth_out ();
      exit (EXIT_FAILURE);
   }

   /* If input filename is specified, open it. */
   if (infile[0] != '\0') {
      if ((infp = fopen (infile, "r")) == NULL) {
         fprintf (stderr, "ERROR: can't open %s for input.\n\n", infile);
         exit (EXIT_FAILURE);
      }
   }

   /* If output filename is specified, open it. */
   if (outfile[0] != '\0') {
      if ((outfp = fopen (outfile, "w")) == NULL) {
         fprintf (stderr, "ERROR: can't open %s for output.\n\n", outfile);
         exit (EXIT_FAILURE);
      }
   }

   /* If a map file was specified, read it and parse it. */
   if (map_file[0] != '\0') {
      if ((mapfp = fopen (map_file, "r")) == NULL) {
         fprintf (stderr, "ERROR - cannot open map file %s\n", mapfile);
         exit (EXIT_FAILURE);
      }

      for (n_entries = 0;
           fgets (linein, MAXSTRING, mapfp) != NULL;
           n_entries++);

      nbytes = ftell (mapfp);

      map = malloc (nbytes * sizeof (char));  /* allocate map array */
      rewind (mapfp);
      fread (map, sizeof (char), nbytes, mapfp);  /* read entire map file */

      fclose (mapfp);

      orig_map = malloc (n_entries * sizeof (char *));
      new_map  = malloc (n_entries * sizeof (char *));

      n_mapped = parseth_map (map, nbytes, orig_map, new_map);
   }  /* if (map_file != NULL) */

   /* Begin output with UTF-8 Byte Order Mark if requested. */
   if (bom_out) fprintf (outfp, "\357\273\277");

   /* Transform each input line as per specified options. */
   while (fgets (instring, MAXSTRING, infp) != NULL) {
      /* If there is a special "\" sequence, handle it first. */
      if ((back_pat = index (instring, '\\')) != NULL) {
         replaceth_back (instring);
      }
      /* If there is a mapped entry for words, replace them first */
      if (n_mapped > 0)
         replaceth_map (instring, orig_map, new_map, n_mapped);
      if (use_mask & USE_ASH )       replaceth_ae (instring);
      if (use_mask & USE_ETHEL)      replaceth_oe (instring);
      if (use_mask & USE_YOGH)       replaceth_gh (instring);
      /* Replace "gh" before changing all 'g's to insular 'g's. */
      if (!(use_mask & USE_INSULAR) &&
           (use_mask & USE_LONG_S))  replaceth_s  (instring);
      if (use_mask & USE_WYNN)       replaceth_w  (instring);
      if (use_mask & USE_ETH || use_mask & USE_THORN) {
         replaceth_th (instring, use_mask);
         replaceth_ye_old (instring, use_mask);
      }
      if (use_mask & USE_INSULAR)    replaceth_insular  (instring);

      fprintf (outfp, "%s", instring);
   }

   exit (EXIT_SUCCESS);
}


/*
   Parse command line arguments.
*/
void parseth_args (int argc, char **argv, int *printeth_help,
                   unsigned *use_mask, char *infile, char *outfile,
                   int *bom_out, char *map_file) {

   int i;  /* loop variable */

   void version_out ();

   infile[0] = outfile[0] = '\0';  /* No input/output file names given yet */

   /* Parse options */
   for (i = 1; i < argc; i++) {
      if (argv[i][0] == '-') {
         switch (argv[i][1]) {
            /*
               Version number request; print on stdout and then exit.
            */
            case '-': if (strcmp (argv[i], "--version") == 0) {
                         version_out ();
                         exit (EXIT_SUCCESS);
                      }
                      else {
                         *printeth_help = 1;
                      }
                      break;
            /*
               General I/O parameters, plus "-oe".
            */
            case 'i': /* Don't print insular letters, U+A779..U+A787 */
                      if (strcmp (argv[i], "-insular") == 0) {
                         *use_mask &= ~USE_INSULAR;
                      }
                      /* input filename follows 'i' as next argument */
                      else if (strcmp (argv[i], "-i") == 0) {
                         strncpy (infile, argv[++i], MAXSTRING);
                      }
                      else {
                         *printeth_help = 1;
                      }
                      break;
            case 'o': if (strcmp (argv[i], "-oe") == 0) {
                         *use_mask &= !USE_ETHEL;
                      }
                      else if (strcmp (argv[i], "-o") == 0) {
                         /* output filename follows 'o' as next argument */
                         strncpy (outfile, argv[++i], MAXSTRING);
                      }
                      else {
                         *printeth_help = 1;
                      }
                      break;
            case 'b': /* start file with UTF-8 Byte Order Mark */
                      *bom_out = 1;
                      break;
            case 'm': if (strcmp (argv[i], "-map") == 0) {  /* map filename */
                         strncpy (map_file, argv[++i], MAXSTRING);
                      }
                      else {
                         *printeth_help = 1;
                      }
                      break;
            /*
               Letter options to disable.
            */
            case 'a': if (strcmp (argv[i], "-all") == 0) {
                         *use_mask = 0x00;  /* remove all mask options */
                      }
                      else if (strcmp (argv[i], "-ae")  == 0 ||
                               strcmp (argv[i], "-ash") == 0) {
                         *use_mask &= ~USE_ASH;  /* Don't convert "ae" */
                      }
                      else {
                         *printeth_help = 1;
                      }
                      break;
            case 'e': if (strcmp (argv[i], "-ethel") == 0) {
                         *use_mask &= ~USE_ETHEL;
                      }
                      else if (strcmp (argv[i], "-eth") == 0) {
                         *use_mask &= ~USE_ETH;  /* Don't convert "th" to eth */
                      }
                      else {
                         *printeth_help = 1;
                      }
                      break;
            case 'g': /* Don't convert "gh" to yogh */
                      if (strcmp (argv[i], "-gh") == 0) {
                         *use_mask &= ~USE_YOGH;
                      }
                      else {
                         *printeth_help = 1;
                      }
                      break;
            case 's': *use_mask &= ~USE_LONG_S;
                      break;
            case 't': if (strcmp (argv[i], "-thorn") == 0) {
                         /* Don't convert "th" to thorn */
                         *use_mask &= ~USE_THORN;
                      }
                      else if (strcmp (argv[i], "-th") == 0) {
                         /* Don't convert "th" to thorn or eth */
                         *use_mask &= ~USE_TH;
                      }
                      else {
                         *printeth_help = 1;
                      }
                      break;
            case 'w': if (strcmp (argv[i], "-w")    == 0 ||
                          strcmp (argv[i], "-wynn") == 0) {
                         /* Don't convert "w" to wynn */
                         *use_mask &= ~USE_WYNN;
                      }
                      else {
                         *printeth_help = 1;
                      }
                      break;
            default:  *printeth_help = 1;
                      break;
         }
      }
      else if (argv[i][0] == '+') {
         switch (argv[i][1]) {
            case 'a': if (strcmp (argv[i], "+all") == 0) {
                         *use_mask = USE_ALL;  /* Do all conversions */
                      }
                      else if (strcmp (argv[i], "+ae")  == 0 ||
                               strcmp (argv[i], "+ash") == 0) {
                         *use_mask |= USE_ASH;  /* Convert "ae" to ash */
                      }
                      else {
                         *printeth_help = 1;
                      }
                      break;
            case 'e': 
                      if (strcmp (argv[i], "+ethel") == 0) {
                         *use_mask |= USE_ETHEL;  /* Convert "oe" to ethel */
                      }
                      else if (strcmp (argv[i], "+eth") == 0) {
                         *use_mask |= USE_ETH;  /* Convert "th" to eth */
                      }
                      else {
                         *printeth_help = 1;
                      }
                      break;
            case 'g': if (strcmp (argv[i], "+gh") == 0) {
                         *use_mask |= USE_YOGH; /* Convert "gh" to yogh */
                      }
                      else {
                         *printeth_help = 1;
                      }
                      break;
            case 'i': /* Print insular letters, U+A779..U+A787 */
                      if (strcmp (argv[i], "+insular") == 0) {
                         *use_mask |= USE_INSULAR; 
                      }
                      else {
                         *printeth_help = 1;
                      }
                      break;
            case 's': *use_mask |= USE_LONG_S; /* Convert "s" to long s */
                      break;
            case 't': if (strcmp (argv[i], "+thorn") == 0) {
                         /* Convert "th" to thorn */
                         *use_mask |= USE_THORN;
                      }
                      else if (strcmp (argv[i], "+th") == 0) {
                         /* Convert "th" to thorn or eth */
                         *use_mask |= USE_TH;
                      }
                      else {
                         *printeth_help = 1;
                      }
                      break;
            case 'w': if (strcmp (argv[i], "+w")    == 0 ||
                          strcmp (argv[i], "+wynn") == 0) {
                         /* Don't convert "w" to wynn */
                         *use_mask |= USE_WYNN;
                      }
                      else {
                         *printeth_help = 1;
                      }
                      break;
            default:  *printeth_help = 1;
                      break;
         }
      }
   }

   return;
}


/*
   Printeth the version number on stdout and exiteth.
*/
void version_out () {

   fprintf (stdout, "olden %s\n", VERSION);
   fprintf (stdout, "Copyright (C) 2019 Paul Hardy <unifoundry@unifoundry.com>\n");
   fprintf (stdout, "License GPLv2+: GNU GPL version 2 or later <http://gnu.org/licenses/gpl.html>\n");
   fprintf (stdout, "This is free software: you are free to change and redistribute it.\n");
   fprintf (stdout, "There is NO WARRANTY, to the extent permitted by law.\n");

   return;
}


/*
   Printeth a help message with options.
*/
void helpeth_out () {

   fprintf (stderr, "\nSyntax:  olden [options]\n");
   fprintf (stderr, "where options are:\n\n");
   fprintf (stderr, "--version    print program information on stdout and then exit\n");
   fprintf (stderr, "+all         enable all conversions (the default).\n");
   fprintf (stderr, "-all         disable all conversions.\n");
   fprintf (stderr, "+ae, +ash    convert \"ae\" to ash.\n");
   fprintf (stderr, "-ae, -ash    don't convert \"ae\" to ash.\n");
   fprintf (stderr, "+oe, +ethel  convert \"oe\" to ethel.\n");
   fprintf (stderr, "-oe, -ethel  don't convert \"oe\" to ethel.\n");
   fprintf (stderr, "+gh, +yogh   convert \"gh\" to yogh.\n");
   fprintf (stderr, "-gh, -yogh   don't convert \"gh\" to yogh.\n");
   fprintf (stderr, "+insular     convert letters d, f, g, r, s, and t to insular forms.\n");
   fprintf (stderr, "-insular     don't convert letters d, f, g, r, s, and t to insular forms.\n");
   fprintf (stderr, "+s           convert non-final 's' to long s.\n");
   fprintf (stderr, "-s           don't convert non-final 's' to long s.\n");
   fprintf (stderr, "+thorn       convert \"th\" to thorn.\n");
   fprintf (stderr, "                Note: also converts voiced \"th\"\n");
   fprintf (stderr, "                if eth conversion is not enabled.\n");
   fprintf (stderr, "-thorn       don't convert \"th\" to thorn.\n");
   fprintf (stderr, "+eth         convert voiced \"th\" to eth.\n");
   fprintf (stderr, "                Note: also converts 'Y' to eth in\n");
   fprintf (stderr, "                strings such as \"Ye Old\".\n");
   fprintf (stderr, "-eth         don't convert voiced \"th\" to eth.\n");
   fprintf (stderr, "+th          convert \"th\" to thorn and eth.\n");
   fprintf (stderr, "-th          don't convert \"th\".\n");
   fprintf (stderr, "+w, +wynn    convert 'w' to wynn.\n");
   fprintf (stderr, "-w, -wynn    don't convert 'w' to wynn.\n");
   fprintf (stderr, "\n");
   fprintf (stderr, "-i filename  open file \"filename\" for input\n");
   fprintf (stderr, "-o filename  open file \"filename\" for output\n");
   fprintf (stderr, "\n");
   fprintf (stderr, "Example:  olden -all +s\n\n");

   return;
}


/*
   Replaceth "@" codes with Unicode characters they represent.
*/
void replaceth_back (char *instring) {
   int i, j;  /* loop variables */
   char tempstring[MAXSTRING];


   i = j = 0;
   while (instring[i] != '\0') {
      if (instring[i] == '\\') {
         switch (instring[i+1]) {
            case '`': if (instring[i+2] == '{' && instring[i+4] == '}') {
                         /* grave accent */
                         if (instring[i+3] == 'i') {
                            /* change "i" to dotless "i" */
                            tempstring[j++] = 0304;
                            tempstring[j++] = 0261;
                         }
                         else if (instring[i+3] == 'j') {
                            /* change "j" to dotless "j" */
                            tempstring[j++] = 0310;
                            tempstring[j++] = 0267;
                         }
                         else {
                            tempstring[j++] = instring[i+3];
                         }
                         tempstring[j++] = 0314;
                         tempstring[j++] = 0200;
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case '\'': if (instring[i+2] == '{' && instring[i+4] == '}') {
                         /* acute accent */
                         if (instring[i+3] == 'i') {
                            /* change "i" to dotless "i" */
                            tempstring[j++] = 0304;
                            tempstring[j++] = 0261;
                         }
                         else if (instring[i+3] == 'j') {
                            /* change "j" to dotless "j" */
                            tempstring[j++] = 0310;
                            tempstring[j++] = 0267;
                         }
                         else {
                            tempstring[j++] = instring[i+3];
                         }
                         tempstring[j++] = 0314;
                         tempstring[j++] = 0201;
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case '^': if (instring[i+2] == '{' && instring[i+4] == '}') {
                         /* circumflex */
                         if (instring[i+3] == 'i') {
                            /* change "i" to dotless "i" */
                            tempstring[j++] = 0304;
                            tempstring[j++] = 0261;
                         }
                         else if (instring[i+3] == 'j') {
                            /* change "j" to dotless "j" */
                            tempstring[j++] = 0310;
                            tempstring[j++] = 0267;
                         }
                         else {
                            tempstring[j++] = instring[i+3];
                         }
                         tempstring[j++] = 0314;
                         tempstring[j++] = 0202;
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case '~': if (instring[i+2] == '{' && instring[i+4] == '}') {
                         /* tilde */
                         if (instring[i+3] == 'i') {
                            /* change "i" to dotless "i" */
                            tempstring[j++] = 0304;
                            tempstring[j++] = 0261;
                         }
                         else if (instring[i+3] == 'j') {
                            /* change "j" to dotless "j" */
                            tempstring[j++] = 0310;
                            tempstring[j++] = 0267;
                         }
                         else {
                            tempstring[j++] = instring[i+3];
                         }
                         tempstring[j++] = 0314;
                         tempstring[j++] = 0203;
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case '"': if (instring[i+2] == '{' && instring[i+4] == '}') {
                         /* dieresis, umlaut */
                         if (instring[i+3] == 'i') {
                            /* change "i" to dotless "i" */
                            tempstring[j++] = 0304;
                            tempstring[j++] = 0261;
                         }
                         else if (instring[i+3] == 'j') {
                            /* change "j" to dotless "j" */
                            tempstring[j++] = 0310;
                            tempstring[j++] = 0267;
                         }
                         else {
                            tempstring[j++] = instring[i+3];
                         }
                         tempstring[j++] = 0314;
                         tempstring[j++] = 0210;
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case '=': if (instring[i+2] == '{' && instring[i+4] == '}') {
                         /* macron */
                         if (instring[i+3] == 'i') {
                            /* change "i" to dotless "i" */
                            tempstring[j++] = 0304;
                            tempstring[j++] = 0261;
                         }
                         else if (instring[i+3] == 'j') {
                            /* change "j" to dotless "j" */
                            tempstring[j++] = 0310;
                            tempstring[j++] = 0267;
                         }
                         else {
                            tempstring[j++] = instring[i+3];
                         }
                         tempstring[j++] = 0314;
                         tempstring[j++] = 0204;
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case ',': if (instring[i+2] == '{' && instring[i+4] == '}') {
                         /* cedilla */
                         if (instring[i+3] == 'C') {  /* "C" & cedilla */
                            tempstring[j++] = 0303;
                            tempstring[j++] = 0207;
                         }
                         else if (instring[i+3] == 'c') {  /* "c" & cedilla */
                            tempstring[j++] = 0303;
                            tempstring[j++] = 0247;
                         }
                         else {
                            tempstring[j++] = instring[i+3];
                            tempstring[j++] = 0314;
                            tempstring[j++] = 0210;
                         }
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case '.': if (instring[i+2] == '{' && instring[i+4] == '}') {
                         /* dot over letter */
                         if (instring[i+3] == 'i') {
                            /* change "i" to dotless "i" */
                            tempstring[j++] = 0304;
                            tempstring[j++] = 0261;
                         }
                         else if (instring[i+3] == 'j') {
                            /* change "j" to dotless "j" */
                            tempstring[j++] = 0310;
                            tempstring[j++] = 0267;
                         }
                         else {
                            tempstring[j++] = instring[i+3];
                         }
                         tempstring[j++] = 0314;
                         tempstring[j++] = 0207;
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'A': if (strncmp (&instring[i+1], "AA{}", 4) == 0) {
                         /* "A" with circle */
                         tempstring[j++] = 0303;
                         tempstring[j++] = 0205;
                         i += 5;
                      }
                      else if (strncmp (&instring[i+1], "AE{}", 4) == 0) {
                         /* "AE" ligature */
                         tempstring[j++] = 0303;
                         tempstring[j++] = 0206;
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'a': if (strncmp (&instring[i+1], "aa{}", 4) == 0) {
                         /* "a" with circle */
                         tempstring[j++] = 0303;
                         tempstring[j++] = 0245;
                         i += 5;
                      }
                      else if (strncmp (&instring[i+1], "ae{}", 4) == 0) {
                         /* "ae" ligature */
                         tempstring[j++] = 0303;
                         tempstring[j++] = 0246;
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'b': if (instring[i+2] == '{' && instring[i+4] == '}') {
                         /* bar under letter */
                         tempstring[j++] = instring[i+3];
                         tempstring[j++] = 0314;
                         tempstring[j++] = 0261;
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'c': if (instring[i+2] == '{' && instring[i+4] == '}') {
                         if (instring[i+3] == 'C') {
                            /* "C" with cedilla */
                            tempstring[j++] = 0303;
                            tempstring[j++] = 0207;
                         }
                         else if (instring[i+3] == 'c') {
                            /* "c" with cedilla */
                            tempstring[j++] = 0303;
                            tempstring[j++] = 0247;
                         }
                         else {
                            /* cedilla */
                            tempstring[j++] = instring[i+3];
                            tempstring[j++] = 0314;
                            tempstring[j++] = 0247;
                         }
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'D': if (strncmp (&instring[i+1], "DH{}", 4) == 0) {
                         /* capital eth */
                         tempstring[j++] = 0303;
                         tempstring[j++] = 0220;
                         i += 5;
                      }
                      else if (strncmp (&instring[i+1], "DJ{}", 4) == 0) {
                         /* capital "D" with stroke */
                         tempstring[j++] = 0304;
                         tempstring[j++] = 0220;
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'd': if (strncmp (&instring[i+1], "dh{}", 4) == 0) {
                         /* lowercase eth */
                         tempstring[j++] = 0303;
                         tempstring[j++] = 0260;
                         i += 5;
                      }
                      else if (strncmp (&instring[i+1], "dj{}", 4) == 0) {
                         /* lowercase "D" with stroke */
                         tempstring[j++] = 0304;
                         tempstring[j++] = 0221;
                         i += 5;
                      }
                      else if (instring[i+2] == '{' && instring[i+4] == '}') {
                         /* dot under letter */
                         tempstring[j++] = instring[i+3];
                         tempstring[j++] = 0314;
                         tempstring[j++] = 0243;
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'G': if (instring[i+2] == '{' && instring[i+4] == '}') {
                         /* double grave accent */
                         if (instring[i+3] == 'i') {
                            /* change "i" to dotless "i" */
                            tempstring[j++] = 0304;
                            tempstring[j++] = 0261;
                         }
                         else if (instring[i+3] == 'j') {
                            /* change "j" to dotless "j" */
                            tempstring[j++] = 0310;
                            tempstring[j++] = 0267;
                         }
                         else {
                            tempstring[j++] = instring[i+3];
                         }
                         tempstring[j++] = 0314;
                         tempstring[j++] = 0217;
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'H': if (instring[i+2] == '{' && instring[i+4] == '}') {
                         /* double acute accent */
                         if (instring[i+3] == 'i') {
                            /* change "i" to dotless "i" */
                            tempstring[j++] = 0304;
                            tempstring[j++] = 0261;
                         }
                         else if (instring[i+3] == 'j') {
                            /* change "j" to dotless "j" */
                            tempstring[j++] = 0310;
                            tempstring[j++] = 0267;
                         }
                         else {
                            tempstring[j++] = instring[i+3];
                         }
                         tempstring[j++] = 0314;
                         tempstring[j++] = 0213;
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'I': if (strncmp (&instring[i+1], "I{}", 3) == 0) {
                         tempstring[j++] = 'I';
                         i += 4;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'i': if (strncmp (&instring[i+1], "i{}", 3) == 0) {
                         /* dotless "i" */
                         tempstring[j++] = 0304;
                         tempstring[j++] = 0261;
                         i += 4;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'J': if (strncmp (&instring[i+1], "J{}", 3) == 0) {
                         tempstring[j++] = 'J';
                         i += 4;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'j': if (strncmp (&instring[i+1], "j{}", 3) == 0) {
                         /* dotless "i" */
                         tempstring[j++] = 0310;
                         tempstring[j++] = 0267;
                         i += 4;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'k': if (instring[i+2] == '{' && instring[i+4] == '}') {
                         /* hook under letter, ogonek */
                         if (instring[i+3] == 'E') {  /* Saxon letter */
                            tempstring[j++] = 0304;
                            tempstring[j++] = 0230;
                         }
                         else if (instring[i+3] == 'e') {  /* Saxon letter */
                            tempstring[j++] = 0304;
                            tempstring[j++] = 0231;
                         }
                         else {
                            tempstring[j++] = instring[i+3];
                            tempstring[j++] = 0314;
                            tempstring[j++] = 0250;
                         }
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'L': if (strncmp (&instring[i+1], "L{}", 3) == 0) {
                         /* "L" with stroke */
                         tempstring[j++] = 0305;
                         tempstring[j++] = 0201;
                         i += 4;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'l': if (strncmp (&instring[i+1], "l{}", 3) == 0) {
                         /* "l" with stroke */
                         tempstring[j++] = 0305;
                         tempstring[j++] = 0202;
                         i += 4;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'N': if (strncmp (&instring[i+1], "NG{}", 4) == 0) {
                         /* "N" with hook */
                         tempstring[j++] = 0305;
                         tempstring[j++] = 0212;
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'n': if (strncmp (&instring[i+1], "ng{}", 4) == 0) {
                         /* "n" with hook */
                         tempstring[j++] = 0305;
                         tempstring[j++] = 0213;
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'O': if (strncmp (&instring[i+1], "O{}", 3) == 0) {
                         /* "O" with slash */
                         tempstring[j++] = 0303;
                         tempstring[j++] = 0230;
                         i += 4;
                      }
                      else if (strncmp (&instring[i+1], "OE{}", 4) == 0) {
                         /* "OE" ligature */
                         tempstring[j++] = 0305;
                         tempstring[j++] = 0222;
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'o': if (strncmp (&instring[i+1], "o{}", 3) == 0) {
                         /* "o" with slash */
                         tempstring[j++] = 0303;
                         tempstring[j++] = 0270;
                         i += 4;
                      }
                      else if (strncmp (&instring[i+1], "oe{}", 4) == 0) {
                         /* "oe" ligature */
                         tempstring[j++] = 0305;
                         tempstring[j++] = 0223;
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'r': if (instring[i+2] == '{' && instring[i+4] == '}') {
                         /* ring over letter */
                         if (instring[i+3] == 'i') {
                            /* change "i" to dotless "i" */
                            tempstring[j++] = 0304;
                            tempstring[j++] = 0261;
                         }
                         else if (instring[i+3] == 'j') {
                            /* change "j" to dotless "j" */
                            tempstring[j++] = 0310;
                            tempstring[j++] = 0267;
                         }
                         else {
                            tempstring[j++] = instring[i+3];
                         }
                         tempstring[j++] = 0314;
                         tempstring[j++] = 0212;
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'S': if (strncmp (&instring[i+1], "SS{}", 4) == 0) {
                         tempstring[j++] = 'S';
                         tempstring[j++] = 'S';
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 's': if (strncmp (&instring[i+1], "ss{}", 4) == 0) {
                         /* German double "s" */
                         tempstring[j++] = 0303;
                         tempstring[j++] = 0237;
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'T': if (strncmp (&instring[i+1], "TH{}", 4) == 0) {
                         /* capital thorn */
                         tempstring[j++] = 0303;
                         tempstring[j++] = 0236;
                         i += 5;
                      }
                      else if (strncmp (&instring[i+1], "Thaet{}", 7) == 0) {
                         tempstring[j++] = 0352;
                         tempstring[j++] = 0235;
                         tempstring[j++] = 0244;
                         i += 8;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 't': if (strncmp (&instring[i+1], "th{}", 4) == 0) {
                         /* lowercase thorn */
                         tempstring[j++] = 0303;
                         tempstring[j++] = 0276;
                         i += 5;
                      }
                      else if (strncmp (&instring[i+1], "thaet{}", 7) == 0) {
                         tempstring[j++] = 0352;
                         tempstring[j++] = 0235;
                         tempstring[j++] = 0245;
                         i += 8;
                      }
                      else if (instring[i+2] == '{' && instring[i+5] == '}') {
                         /* tie over two letters */
                         if (instring[i+3] == 'i') {
                            /* change "i" to dotless "i" */
                            tempstring[j++] = 0304;
                            tempstring[j++] = 0261;
                         }
                         else if (instring[i+3] == 'j') {
                            /* change "j" to dotless "j" */
                            tempstring[j++] = 0310;
                            tempstring[j++] = 0267;
                         }
                         else {
                            tempstring[j++] = instring[i+3];
                         }
                         tempstring[j++] = 0315;
                         tempstring[j++] = 0241;
                         if (instring[i+4] == 'i') {
                            /* change "i" to dotless "i" */
                            tempstring[j++] = 0304;
                            tempstring[j++] = 0261;
                         }
                         else if (instring[i+4] == 'j') {
                            /* change "j" to dotless "j" */
                            tempstring[j++] = 0310;
                            tempstring[j++] = 0267;
                         }
                         else {
                            tempstring[j++] = instring[i+4];
                         }
                         i += 6;
                      }
                      else if (strncmp (&instring[i+1], "text", 4) == 0) {
                         if (strncmp (&instring[i+5], "exclamdown{}", 12) == 0) {
                            /* inverted exclamation mark */
                            tempstring[j++] = 0302;
                            tempstring[j++] = 0241;
                            i += 17;
                         }
                         else if (strncmp (&instring[i+5], "questiondown{}", 14) == 0) {
                            /* inverted question mark */
                            tempstring[j++] = 0302;
                            tempstring[j++] = 0277;
                            i += 19;
                         }
                         else if (strncmp (&instring[i+5], "ordfeminine{}", 13) == 0) {
                            /* Spanish feminime ordinal */
                            tempstring[j++] = 0302;
                            tempstring[j++] = 0252;
                            i += 18;
                            }
                         else if (strncmp (&instring[i+5], "ordmasculine{}", 14) == 0) {
                            /* Spanish masculine ordinal */
                            tempstring[j++] = 0302;
                            tempstring[j++] = 0272;
                            i += 19;
                            }
                         else {
                            tempstring[j++] = instring[i++];
                         }
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'U': if (instring[i+2] == '{' && instring[i+4] == '}') {
                         /* combining double vertical line above */
                         if (instring[i+3] == 'i') {
                            /* change "i" to dotless "i" */
                            tempstring[j++] = 0304;
                            tempstring[j++] = 0261;
                         }
                         else if (instring[i+3] == 'j') {
                            /* change "j" to dotless "j" */
                            tempstring[j++] = 0310;
                            tempstring[j++] = 0267;
                         }
                         else {
                            tempstring[j++] = instring[i+3];
                         }
                         tempstring[j++] = 0314;
                         tempstring[j++] = 0216;
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'u': if (instring[i+2] == '{' && instring[i+4] == '}') {
                         /* breve mark over letter */
                         if (instring[i+3] == 'i') {
                            /* change "i" to dotless "i" */
                            tempstring[j++] = 0304;
                            tempstring[j++] = 0261;
                         }
                         else if (instring[i+3] == 'j') {
                            /* change "j" to dotless "j" */
                            tempstring[j++] = 0310;
                            tempstring[j++] = 0267;
                         }
                         else {
                            tempstring[j++] = instring[i+3];
                         }
                         tempstring[j++] = 0314;
                         tempstring[j++] = 0206;
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
                      break;
            case 'v': if (instring[i+2] == '{' && instring[i+4] == '}') {
                         /* caron over letter */
                         if (instring[i+3] == 'i') {
                            /* change "i" to dotless "i" */
                            tempstring[j++] = 0304;
                            tempstring[j++] = 0261;
                         }
                         else if (instring[i+3] == 'j') {
                            /* change "j" to dotless "j" */
                            tempstring[j++] = 0310;
                            tempstring[j++] = 0267;
                         }
                         else {
                            tempstring[j++] = instring[i+3];
                         }
                         tempstring[j++] = 0314;
                         tempstring[j++] = 0214;
                         i += 5;
                      }
                      else {
                         tempstring[j++] = instring[i++];
                      }
            default:  tempstring[j++] = instring[i++];
                      break;
         }
      }  /* if (instring[i] == '\\') */
      else {
         tempstring[j++] = instring[i++];
      }
   }  /* while (instring[i] != '\0') */

   tempstring[j] = '\0';
   strncpy (instring, tempstring, MAXSTRING);

   return;
}


/*
   Replaceth "AE" with ash.
*/
void replaceth_ae (char *sstring) {
   char *cstart;  /* start of character string within a word */
   char *inptr;   /* pointer into instring */
   int   use_upper; /* whether new letter is upper- or lower-case */

   char *strcasestr (const char *big, const char *little);
   size_t strcspn (const char *, const char *);


   inptr = sstring;

   /*
      Handle conversion of "AE".
   */
   while (*inptr != '\0') {
      cstart = strcasestr (inptr, "AE");

      if (cstart != NULL) {
         use_upper = isupper (*cstart);

         if (use_upper) {  /* U+00C6 */
            cstart[0] = '\303';
            cstart[1] = '\206';
         }
         else {  /* U+00E6 */
            cstart[0] = '\303';
            cstart[1] = '\246';
         }

         inptr = cstart + 2;  /* point past "AE" */
      }
      else {
         inptr = &sstring [strlen (sstring)]; /* point to end of sstring */
      }

   }  /* while (*inptr != '\0') */

   return;
}


/*
   Replaceth "OE" with ethel.
*/
void replaceth_oe (char *sstring) {
   char *cstart;  /* start of character string within a word */
   char *inptr;   /* pointer into instring */
   int   use_upper; /* whether new letter is upper- or lower-case */

   char *strcasestr (const char *big, const char *little);
   size_t strcspn (const char *, const char *);


   inptr = sstring;

   /*
      Handle conversion of "OE".
   */
   while (*inptr != '\0') {
      cstart = strcasestr (inptr, "OE");

      if (cstart != NULL) {
         use_upper = isupper (*cstart);

         if (use_upper) {  /* U+0152 */
            cstart[0] = '\305';
            cstart[1] = '\222';
         }
         else {  /* U+0153 */
            cstart[0] = '\305';
            cstart[1] = '\223';
         }

         inptr = cstart + 2;  /* point past "OE" */
      }
      else {
         inptr = &sstring [strlen (sstring)]; /* point to end of sstring */
      }

   }  /* while (*inptr != '\0') */

   return;
}


/*
   Replaceth "GH" with yogh.
*/
void replaceth_gh (char *sstring) {
   char *cstart;  /* start of character string within a word */
   char *inptr;   /* pointer into instring */
   int   use_upper; /* whether new letter is upper- or lower-case */

   char *strcasestr (const char *big, const char *little);
   size_t strcspn (const char *, const char *);


   inptr = sstring;

   /*
      Handle conversion of "GH".
   */
   while (*inptr != '\0') {
      cstart = strcasestr (inptr, "GH");

      if (cstart != NULL) {
         use_upper = isupper (*cstart);

         if (use_upper) {  /* U+021C */
            cstart[0] = '\310';
            cstart[1] = '\234';
         }
         else {  /* U+021D */
            cstart[0] = '\310';
            cstart[1] = '\235';
         }

         inptr = cstart + 2;  /* point past "GH" */
      }
      else {
         inptr = &sstring [strlen (sstring)]; /* point to end of sstring */
      }

   }  /* while (*inptr != '\0') */

   return;
}


/*
   Replaceth d, f, g, r, s, and t with insular forms.
*/
void replaceth_insular (char *sstring) {
   int   i, j;    /* loop variables */
   int   found_insular; /* found at least one insular letter to convert */
   char  newstring[MAXSTRING];  /* temporary string for substitutions */


   found_insular = 0;

   /*
      See if line contains a letter that has an insular form.
      If not, we won't need to copy the line into newstring,
      inserting insular forms, and copying back to sstring.
   */
   for (i = 0; sstring[i] != '\0' && found_insular == 0; i++) {
      if (index ("DdFfGgRrSsTt", sstring[i]) != NULL) {
         found_insular++;
      }
   }

   /*
      Now convert to UTF-8 if any initial or medial "s" was found.
   */
   if (found_insular > 0) {
      for (i = 0, j = 0;
           sstring[i] != '\0' && j < (MAXSTRING - 1);
           i++, j++) {

         switch (sstring[i]) {
            case 'D': newstring[j++] = '\352';  /* U+A779 */
                      newstring[j++] = '\235';
                      newstring[j]   = '\271';
                      break;
            case 'd': newstring[j++] = '\352';  /* U+A77A */
                      newstring[j++] = '\235';
                      newstring[j]   = '\272';
                      break;
            case 'F': newstring[j++] = '\352';  /* U+A77B */
                      newstring[j++] = '\235';
                      newstring[j]   = '\273';
                      break;
            case 'f': newstring[j++] = '\352';  /* U+A77C */
                      newstring[j++] = '\235';
                      newstring[j]   = '\274';
                      break;
            case 'G': newstring[j++] = '\352';  /* U+A77D */
                      newstring[j++] = '\235';
                      newstring[j]   = '\275';
                      break;
            case 'g': newstring[j++] = '\341';  /* U+1D79 */
                      newstring[j++] = '\265';
                      newstring[j]   = '\271';
                      break;
            case 'R': newstring[j++] = '\352';  /* U+A782 */
                      newstring[j++] = '\236';
                      newstring[j]   = '\202';
                      break;
            case 'r': newstring[j++] = '\352';  /* U+A783 */
                      newstring[j++] = '\236';
                      newstring[j]   = '\203';
                      break;
            case 'S': newstring[j++] = '\352';  /* U+A784 */
                      newstring[j++] = '\236';
                      newstring[j]   = '\204';
                      break;
            case 's': newstring[j++] = '\352';  /* U+A785 */
                      newstring[j++] = '\236';
                      newstring[j]   = '\205';
                      break;
            case 'T': newstring[j++] = '\352';  /* U+A786 */
                      newstring[j++] = '\236';
                      newstring[j]   = '\206';
                      break;
            case 't': newstring[j++] = '\352';  /* U+A787 */
                      newstring[j++] = '\236';
                      newstring[j]   = '\207';
                      break;
            default: newstring[j] = sstring[i];
                     break;
         }  /* switch */
      }  /* for */
      newstring[j] = '\0';

      /*
         Copy newstring back to instring, from end to start.
      */
      for ( ; j >= 0; j--) sstring[j] = newstring[j];
   }  /* if */
   

   return;
}


/*
   Replaceth initial and medial "s".
*/
void replaceth_s (char *sstring) {
   int   i, j;    /* loop variables */
   int   found_s; /* found at least one "s" to convert */
   char  newstring[MAXSTRING];  /* temporary string for substitutions */


   found_s = 0;

   /*
      Find all initial and medial "s"s and mark as 0x0E.
   */
   if (sstring [0] == 's') {
      sstring [0] = 0x0E;  /* mark for change below */
      found_s++;
   }

   for (i = 1; sstring[i] != '\0'; i++) {
      if (sstring[i] == 's') {
         /* If not preceded by "s" and not a final "s", mark for change. */
         if ((sstring[i-1] != 0x0E) && isalpha (sstring[i+1])) {
            sstring[i] = 0x0E;
            found_s++;
         }
      }
   }

   /*
      Now convert to UTF-8 if any initial or medial "s" was found.
   */
   if (found_s > 0) {
      for (i = 0, j = 0;
           sstring[i] != '\0' && j < (MAXSTRING - 1);
           i++, j++) {

         if (sstring[i] == 0x0E) {
            newstring[j++] = '\305';  /* U+017F */
            newstring[j]   = '\277';
         }
         else {
            newstring[j] = sstring[i];
         }
      }
      newstring[j] = '\0';

      /*
         Copy newstring back to instring, from end to start.
      */
      for ( ; j >= 0; j--) sstring[j] = newstring[j];
   }
   

   return;
}


/*
   Replaceth "w" with wynn.
*/
void replaceth_w (char *sstring) {
   int   i, j;    /* loop variables */
   int   found_w; /* found at least one "s" to convert */
   char  newstring[MAXSTRING];  /* temporary string for substitutions */


   found_w = 0;

   /*
      Find all "W"s and mark as 0x0E, all "w"s and mark as 0x0F.
   */
   if (sstring [0] == 'W') {
      sstring [0] = 0x0E;  /* mark for change below */
      found_w++;
   }
   else if (sstring [0] == 'w') {
      sstring [0] = 0x0F;  /* mark for change below */
      found_w++;
   }

   for (i = 1; sstring[i] != '\0'; i++) {
      if (sstring[i] == 'W') {
            sstring[i] = 0x0E;
            found_w++;
         }
      else if (sstring[i] == 'w') {
         sstring[i] = 0x0F;
         found_w++;
      }
   }

   /*
      Now convert to UTF-8 if any "W" or "w" was found.
   */
   if (found_w > 0) {
      for (i = 0, j = 0;
           sstring[i] != '\0' && j < (MAXSTRING - 1);
           i++, j++) {

         if (sstring[i] == 0x0E) {
            newstring[j++] = '\307';  /* U+01F7, upper-case wynn */
            newstring[j]   = '\267';
         }
         else if (sstring[i] == 0x0F) {
            newstring[j++] = '\306';  /* U+01BF, lower-case wynn */
            newstring[j]   = '\277';
         }
         else {
            newstring[j] = sstring[i];
         }
      }
      newstring[j] = '\0';

      /*
         Copy newstring back to instring, from end to start.
      */
      for ( ; j >= 0; j--) sstring[j] = newstring[j];
   }
   

   return;
}


/*
   Replaceth "TH" if appropriate.
*/
void replaceth_th (char *sstring, unsigned use_mask) {
   int   i, j;    /* loop variables */
   char *cstart;  /* start of character string within a word */
   char *wstart;  /* start of word we're interested in */
   char *wend;    /* end of word we're interested in */
   char  savechar; /* character to save, to insert '\0' in a string */
   char *inptr;   /* pointer into instring */
   char  newstring [MAXSTRING];  /* converted string to copy back */
   int   use_upper; /* whether new letter is upper- or lower-case */

   int findeth_eth (char *);
   int findeth_t4two (char *);

   char *strcasestr (const char *big, const char *little);
   size_t strcspn (const char *, const char *);


   inptr = sstring;

   /*
      Handle conversion of "TH".
   */
   while (*inptr != '\0') {
      cstart = strcasestr (inptr, "TH");

      if (cstart != NULL) {
         use_upper = isupper (*cstart);

         /* find start of word */
         for (wstart = cstart;
              wstart >= inptr && isalpha (*wstart);
              wstart--);

         if (wstart < inptr || !isalpha (*wstart)) wstart++;

         /* find end of word */
         for (wend = cstart; isalpha (*wend); wend++);

         /* save character after this word to restore later */
         savechar = *wend;
         *wend = '\0';
         wend--;

         /*
            Now the word we want starts at wstart and ends at wend.
         */

         /*
           If not a "th" combination to preserve, change it.  Eth
           and thorn are both two bytes in UTF-8, so we can replace
           "th" with either, in place in the original string.
         */
         if (findeth_t4two (wstart) < 0) {
            /* Replace with eth? */
            if ((use_mask & USE_ETH) && findeth_eth (wstart) >= 0) {
               if (use_upper) {  /* U+00D0, upper-case eth */
                  cstart[0] = '\303';
                  cstart[1] = '\220';
               }
               else {
                  cstart[0] = '\303';  /* U+00F0, lower-case eth */
                  cstart[1] = '\260';
               }
               /* otherwise, leave as "th" */
            }
            /* Don't replace with eth; replace with thorn */
            else if (use_mask & USE_THORN) {
               if (use_upper) {
                  cstart[0] = '\303';  /* U+00DE */
                  cstart[1] = '\236';
               }
               else {  /* U+00FE */
                  cstart[0] = '\303';
                  cstart[1] = '\276';
               }
               /* otherwise, leave as "th" */
            }
         }

         /* restore character that was overwritten with NULL */
         wend++;
         *wend = savechar;

         inptr = wend;
      }
      else {
         inptr = &sstring [strlen (sstring)]; /* point to end of sstring */
      }

   }  /* while (*inptr != '\0') */

   return;
}


/*
   Replaceth mapped words if appropriate.
*/
void replaceth_map (char *sstring, char **orig_map, char **new_map,
                    int n_mapped) {
   int   i, j, k;  /* loop variables */
   char *cstart;   /* start of character string within a word */
   char  wstart;   /* start of word we're interested in */
   char  wend;     /* end of word we're interested in */
   char  savechar; /* character to save, to insert '\0' in a string */
   char  newstring [MAXSTRING];  /* converted string to copy back */
   int   use_upper; /* whether new letter is upper- or lower-case */
   int   word_loc;  /* location of word entry in map arrays */

   int binsearch (char *sstring, char **sarray, int n);

   char *strcasestr (const char *big, const char *little);
   size_t strcspn (const char *, const char *);


   /*
      Handle conversion of "TH".

      i is the index into the input string, sstring.
      j is the index into the output sttring, newstring.
   */
   i = j = 0;
   while (sstring[i] != '\0') {
      /* skip over non-alphabetical characters */
      while (!isalpha (sstring[i]) && sstring[i] != '\0') {
         newstring[j++] = sstring[i++];
      }

      /*
         Look up one word at a time to see if it is one
         with a mapping to a new word.
      */
      if (isalpha (sstring[i])) {
         wstart = i;
         wend   = i + 1;
         while (isalpha (sstring[wend])) wend++;  /* find end of this word */
         /* save character after this word to restore later */
         savechar = sstring[wend];
         sstring[wend] = '\0';

         /*
            If word starts with an upper case letter,
            make it lower-case because the mapping list
            is all lower-case.  We'll restore the upper-case
            letter later.
         */
         use_upper  = isupper (sstring[i]);
         sstring[i] = tolower (sstring[i]);
         word_loc   = binsearch (&sstring[i], orig_map, n_mapped);
         /*
            If a mapping exists, copy it.
         */
         if (word_loc >= 0) {
            for (k = 0; new_map[word_loc][k] != '\0'; j++, k++) {
               newstring[j] = new_map[word_loc][k];
               /* Restore upper-case first letter if it was present */
               if (k == 0 && use_upper) {
                  if (new_map[word_loc][0] < 0x7F) {  /* ASCII letter */
                     newstring[j] = toupper (newstring[j]);
                  }
                  else {  /* Handle eth and thorn */
                     if ((unsigned char)new_map[word_loc][0] == 0303) {
                        if ((unsigned char)new_map[word_loc][1] == 0260) {
                           newstring[++j] = 0220;  /* uppercase eth */
                           k++;
                        }
                        else if ((unsigned char)new_map[word_loc][1] == 0276) {
                           newstring[++j] = 0236;  /* uppercase thorn */
                           k++;
                        }
                     }
                  }
               }
            }
            while (sstring[i] != '\0') i++;  /* skip over this input word */
         }
         else {  /* no map for this word; just copy it */
            /* Restore upper-case first letter if it was present */
            if (use_upper) sstring[i] = toupper (sstring[i]);
            while (isalpha (sstring[i])) newstring[j++] = sstring[i++];
         }
         sstring[wend] = savechar;  /* Restore overwritten character */
      }  /* if (sstring[i] != '\0') */
   }  /* while (sstring[i] != '\0') */

   newstring[j]   = '\0';

   /*
      Copy modified string back to input string  for further processing.
   */
   strncpy (sstring, newstring, MAXSTRING);

   return;
}


/*
   Searcheth for input string in eth array.
   Returneth index if found, -1 if not found.
*/
int findeth_eth (char *sstring) {
   int  i;
   char tempstring[MAXSTRING];
   int  retval;  /* return value */

   int binsearch (char *, char **, int);


   /* If first letter is upper-case, convert word to lower-case. */
   if (isupper (sstring[0])) {
      for (i = 0; sstring[i] != '\0'; i++) {
         tempstring [i] = tolower (sstring [i]);
      }
      retval = binsearch (tempstring, eth, N_ETH);
   }
   else {
      retval = binsearch (sstring, eth, N_ETH);
   }

   return (retval);
}


/*
   Searcheth for input string in t4two array, to leave "th" unmodified.
   Returneth index if found, -1 if not found.
*/
int findeth_t4two (char *sstring) {
   int  i;
   char tempstring[MAXSTRING];
   int  retval;  /* return value */

   int binsearch (char *, char **, int);


   /* If first letter is upper-case, convert word to lower-case. */
   if (isupper (sstring[0])) {
      for (i = 0; sstring[i] != '\0'; i++) {
         tempstring [i] = tolower (sstring [i]);
      }
      tempstring[i] = '\0';
      retval = binsearch (tempstring, t4two, N_T4TWO);
   }
   else {
      retval = binsearch (sstring, t4two, N_T4TWO);
   }

   return (retval);
}


/*
   Performeth binary search for a string in an array of strings.

   sstring  -- the search string
   sarray   -- the array of strings to search
   n        -- the number of elements in the search string array

   This function returns the index number in the search string
   if found, or -1 if not found.
*/
int binsearch (char *sstring, char **sarray, int n) {
   int left, mid, right;  /* for binary search in eth array */
   int scomp;   /* value of string comparison, < 0, == 0, or > 0 */
   int retval;  /* return value */

   left = 0;
   right = n - 1;  /* maximum array index in search array */

   do {
      mid = (left + right) >> 1;
      scomp = strcmp (sarray[mid], sstring);
      if (scomp < 0) {
         left = mid + 1;
      }
      else if (scomp > 0) {
         right = mid - 1;
      }
   }  while ((left <= right) && (scomp != 0));

   if (scomp == 0)
      retval = mid;
   else
      retval = -1;

   return (retval);
}


/*
   Replaceth the "Y" in "Ye Old" with eth (or thorn, if eth not being used).
*/
void replaceth_ye_old (char *sstring, unsigned use_mask) {
   int   i, j;    /* loop variables */
   char *cstart;  /* start of character string within a word */
   char *inptr;   /* pointer into instring */
   int   use_upper; /* whether new letter is upper- or lower-case */
   int   found_ye_old;  /* found "Ye Old" or "ye old" string */
   char  newstring[MAXSTRING];  /* temporary replacement string */

   char *strcasestr (const char *big, const char *little);
   size_t strcspn (const char *, const char *);


   found_ye_old = 0;

   inptr = sstring;

   /*
      Handle conversion of "Ye Old[e]".
   */
   while (*inptr != '\0') {
      cstart = strcasestr (inptr, "YE OLD");

      if (cstart != NULL) {
         use_upper = isupper (*cstart);

         /*
            Find all "Ye Old"s and mark "Y" as 0x0E, all "y"s and mark as 0x0F.
         */
         if (*cstart == 'Y') {
            sstring [0] = 0x0E;  /* mark for change below */
            found_ye_old++;
         }
         else {
            sstring [0] = 0x0F;  /* mark for change below */
            found_ye_old++;
         }
      }
      else {
         inptr = &sstring [strlen (sstring)]; /* point to end of sstring */
      }
   }  /* while (*inptr != '\0') */

   /*
      Now convert to UTF-8 if any "Ye Old" or "ye old" was found.
   */
   if (found_ye_old > 0) {
      for (i = 0, j = 0;
           sstring[i] != '\0' && j < (MAXSTRING - 1);
           i++, j++) {

         if (sstring[i] == 0x0E) {
            if (use_mask & USE_ETH) {
               newstring[j++] = '\303';  /* U+00D0, upper-case eth */
               newstring[j]   = '\220';
            }
            else { /* use thorn */
               cstart[0] = '\303';  /* U+00DE, upper-case thorn */
               cstart[1] = '\236';
            }
         }
         else if (sstring[i] == 0x0F) {
            if (use_mask & USE_ETH) {
               newstring[j++] = '\303';  /* U+0F0, , lower-case eth */
               newstring[j]   = '\260';
            }
            else { /* use thorn */
               cstart[0] = '\303';  /* U+00FE, lower-case thorn */
               cstart[1] = '\276';
            }
         }
         else {
            newstring[j] = sstring[i];
         }
         /* Next 5 characters are "e old", so copy as is. */
         newstring[++j] = sstring[++i];
         newstring[++j] = sstring[++i];
         newstring[++j] = sstring[++i];
         newstring[++j] = sstring[++i];
         newstring[++j] = sstring[++i];
      }
      newstring[j] = '\0';

      /*
         Copy newstring back to instring, from end to start.
      */
      for ( ; j >= 0; j--) sstring[j] = newstring[j];
   }
   

   return;
}


/*
   parseth_map - parse a file of "<ASCII-word>\t<new-word>\n" lines.

      map       - the entire map file in a character array.
      nbytes    - number of bytes in the map array.
      orig_map - array of the original ASCII words.
      new_map  - array of the corresponding new, replacement words.

   Return Value:

      -1    - map file could not be opened.
       0    - no entries read from file.
      >0    - number of map entries read from map file.
*/
int parseth_map (char *map, long nbytes, char **orig_map, char **new_map) {
   int   i;  /* loop variable */
   int   n_mapped;   /* index of current word being mapped */
   int   retval;     /* return value */


   for (i = n_mapped = 0; i < nbytes; i++) {
      /* Start of a new line with an <original>\t<new>\n entry */
      orig_map[n_mapped] = &map[i];
      while (!isspace (map[i])) i++;  /* skip to end of word */
      do {
         map[i++] = '\0';
      }  while ( isspace (map[i]));  /* skip to start of next word */
      new_map[n_mapped] = &map[i];  /* start of new word to map orig to */
      while (!isspace (map[i])) i++;  /* skip to end of word */
      while (map[i] != '\n')  map[i++] = '\0';
      map[i] = '\0';  /* replace newline with null */
      n_mapped++;
   }
   retval = n_mapped;

return (retval);
}

